import datetime
import functools
import inspect
import itertools
import math
from operator import itemgetter
import random
import re
import sys
from typing import List, Optional, Union

import pytz
from pytz import UnknownTimeZoneError

from ._config import config


class Tokenizer:
    def __init__(self, expression: str):
        self.expression = expression

    def get_next(self, separator: Optional[Union[List[str], str]] = None) -> str:
        if isinstance(separator, list):
            separator_positions_and_lengths = [
                (self.expression.find(s), s)
                for s in separator if s in self.expression
            ]
            if separator_positions_and_lengths:
                sep_pos, separator = min(separator_positions_and_lengths, key=itemgetter(0))
            else:
                sep_pos = -1
        elif separator:
            sep_pos = self.expression.find(separator)
        else:
            sep_pos = -1

        if sep_pos < 0:
            value = self.expression
            self.expression = ''
            return value

        value = self.expression[:sep_pos]
        self.expression = self.expression[sep_pos + len(separator):]
        return value


def parse_file_uri(expr):
    t = Tokenizer(expr)
    scheme = t.get_next('://')
    domain = t.get_next('/')
    last_slash_position = t.expression.rfind('/')
    folder_path = '/' + t.expression[:last_slash_position + 1]
    file_pattern = t.expression[last_slash_position + 1:]

    return scheme, domain, folder_path, file_pattern


def format_file_uri(scheme, domain, *local_path_components):
    return f'{scheme}://{domain}{"/".join(local_path_components)}'


def reservoir_sample_and_size(iterable, k, seed):
    """
    Returns a sample of k items of iterable and its original size
    If the iterable contains less than k items the sample is a list of those items
    Algorithm used is reservoir sampling.

    :rtype list
    """
    random.seed(seed)

    # Put the first k elements in the reservoir.
    reservoir = list(itertools.islice(iterable, k))

    # If we have consumed all the elements, return them. Otherwise do the replacement.
    reservoir_size = len(reservoir)
    if reservoir_size < k:
        return reservoir, reservoir_size

    # If input size > k, continue the sampling process.
    for reservoir_size, item in enumerate(iterable, start=k + 1):
        # There are k elements in the reservoir, and the l-th element has been
        # consumed. It should be chosen with probability k/l. The expression
        # below is a random int chosen uniformly from [0, l)
        replacementIndex = random.randint(0, reservoir_size)
        if replacementIndex < k:
            reservoir[replacementIndex.toInt] = item

    return reservoir, reservoir_size


def compute_weighted_percentiles(weighted_values, number_of_percentiles, key=lambda x: x):
    """
    Compute weighted percentiles from a list of values and weights.

    number_of_percentiles evenly distributed percentiles values will be returned,
    including the 0th (minimal value) and the 100th (maximal value).

    A custom key function can be supplied to customize the sort order

    :type weighted_values: list of tuple
    :type number_of_percentiles: int
    :type key: function

    Examples with 0th, 50th and 100th percentiles:
    >>> compute_weighted_percentiles([(2, 0.2), (1, 0.1), (3, 0.7)], 3)
    [1, 3, 3]
    >>> compute_weighted_percentiles([(1, 10), (2, 20), (3, 20)], 3)
    [1, 2, 3]
    >>> compute_weighted_percentiles([(i, 1) for i in range(1, 101)], 1)
    Traceback (most recent call last):
     ...
    ValueError: number_of_percentiles must be at least 2
    >>> compute_weighted_percentiles([(i, 1) for i in range(1, 101)], 2)
    [1, 100]
    >>> compute_weighted_percentiles([(i, 1) for i in range(1, 101)], 3)
    [1, 50, 100]
    >>> compute_weighted_percentiles([(i, 1) for i in range(1, 101)], 4)
    [1, 34, 67, 100]
    >>> compute_weighted_percentiles([(i, 1) for i in range(1, 101)], 5)
    [1, 25, 50, 75, 100]
    >>> compute_weighted_percentiles([
    ...   ((1, "b"), 10),
    ...   ((2, "c"), 20),
    ...   ((3, "a"), 20)
    ... ], 3, key=lambda row: row[1])
    [(3, 'a'), (1, 'b'), (2, 'c')]
    """
    if number_of_percentiles == 1:
        raise ValueError("number_of_percentiles must be at least 2")

    ordered_values = sorted(weighted_values, key=lambda weighted_value: key(weighted_value[0]))
    total_weight = sum(weight for value, weight in ordered_values)

    bounds = []
    cumulative_weight = 0
    for value, weight in ordered_values:
        cumulative_weight += weight
        while len(bounds) / (number_of_percentiles - 1) <= cumulative_weight / total_weight:
            bounds.append(value)

    return bounds


def get_keyfunc(cols, schema, nulls_are_smaller=False):
    """
    Return a function that maps a row to a tuple of some of its columns values
    """

    def key(row):
        """
        Returns a tuple designed for comparisons based on a row

        Each requested columns is mapped to two columns:
        - The first indicate whether the column value is None or not
        - The second is the column value

        This prevent comparison between None and non-None values
        It also allows to defined how None values should be considered
        """
        values = []
        for col in cols:
            value = col.eval(row, schema)
            values += (
                (value is None) != nulls_are_smaller,
                value
            )
        return tuple(values)

    return key


class MonotonicallyIncreasingIDGenerator:
    def __init__(self, partition_index):
        self.value = partition_index * 8589934592 - 1

    def __iter__(self):
        return self

    def __next__(self):
        self.value += 1
        return self.value


# pylint: disable=W0511
# todo: store random-related utils in a separated module
class XORShiftRandom:
    # pylint: disable=W0511
    # todo: align generated values with the ones in Spark
    def __init__(self, init):
        self.seed = XORShiftRandom.hashSeed(init)
        self.haveNextNextGaussian = False
        self.nextNextGaussian = 0

    def nextValue(self, bits):
        seed = self.seed
        nextSeed = seed ^ (seed << 21)
        nextSeed ^= (nextSeed >> 35)
        nextSeed ^= (nextSeed << 4)
        self.seed = nextSeed
        return int(nextSeed & ((1 << bits) - 1))

    def nextDouble(self):
        a = self.nextValue(26)
        b = self.nextValue(27)
        return ((a << 27) + b) * 1.1102230246251565E-16

    def nextGaussian(self):
        if self.haveNextNextGaussian:
            self.haveNextNextGaussian = False
            return self.nextNextGaussian

        v1 = 0
        v2 = 0
        s = 0
        while not 0 < s < 1:
            v1 = 2.0 * self.nextDouble() - 1
            v2 = 2.0 * self.nextDouble() - 1
            s = v1 * v1 + v2 * v2

        multiplier = math.sqrt(-2 * math.log(s) / s)
        self.nextNextGaussian = v2 * multiplier
        self.haveNextNextGaussian = True
        return v1 * multiplier

    @staticmethod
    def hashSeed(seed):
        as_bytes = seed.to_bytes(8, "big")
        lowBits = MurmurHash3.bytesHash(as_bytes)
        highBits = MurmurHash3.bytesHash(as_bytes, lowBits)
        return (highBits << 32) | (lowBits & 0xFFFFFFFF)


class MurmurHash3:
    @staticmethod
    def bytesHash(data, seed=0x3c074a61):
        length = len(data)
        h = seed

        # Body
        i = 0
        while length >= 4:
            k = data[i + 0] & 0xFF
            k |= (data[i + 1] & 0xFF) << 8
            k |= (data[i + 2] & 0xFF) << 16
            k |= (data[i + 3] & 0xFF) << 24

            h = MurmurHash3.mix(h, k)

            i += 4
            length -= 4

        # Tail
        k = 0
        if length == 3:
            k ^= (data[i + 2] & 0xFF) << 16
        if length >= 2:
            k ^= (data[i + 1] & 0xFF) << 8
        if length >= 1:
            k ^= (data[i + 0] & 0xFF)
            h = MurmurHash3.mixLast(h, k)

        # Finalization
        return MurmurHash3.finalizeHash(h, len(data))

    @staticmethod
    def mix(h, data):
        h = MurmurHash3.mixLast(h, data)
        h = MurmurHash3.rotl(h, 13)
        return h * 5 + 0xe6546b64

    @staticmethod
    def finalizeHash(h, length):
        return MurmurHash3.avalanche(h ^ length)

    @staticmethod
    def avalanche(h):
        h ^= h >> 16
        h *= 0x85ebca6b
        h ^= h >> 13
        h *= 0xc2b2ae35
        h ^= h >> 16
        return h

    @staticmethod
    def mixLast(h, k):
        k *= 0xcc9e2d51
        k = MurmurHash3.rotl(k, 15)
        k *= 0x1b873593

        return h ^ k

    @staticmethod
    def rotl(i, distance):
        return i << distance


def strhash(string):
    """
    Old python hash function as described in PEP 456, excluding prefix, suffix and mask.

    :param string: string to hash
    :return: hash
    """
    if string == "":
        return 0

    x = ord(string[0]) << 7
    for c in string[1:]:
        x = ((1000003 * x) ^ ord(c)) & (1 << 32)
    x = (x ^ len(string))
    return x


def portable_hash(x):
    """
    This function returns consistent hash code for builtin types, especially
    for None and tuple with None.
    The algorithm is similar to that one used by CPython 2.7
    >>> portable_hash(None)
    0
    >>> portable_hash((None, 1)) & 0xffffffff
    219750521
    """

    if x is None:
        return 0
    if isinstance(x, list):
        return portable_hash(tuple(x))
    if isinstance(x, tuple):
        h = 0x345678
        for i in x:
            h ^= portable_hash(i)
            h *= 1000003
            h &= sys.maxsize
        h ^= len(x)
        if h == -1:
            h = -2
        return int(h)
    if isinstance(x, str):
        return strhash(x)
    if isinstance(x, datetime.datetime):
        return portable_hash(x.timetuple())
    return hash(x)


def parse_tz(tz):
    """
    Parse a string referencing a timezone which is either supported by pytz or
    in a GMT+1 or GMT+1:30 format.

    Returns a datetime.tzinfo if it was able to parse the string, None otherwise

    >>> parse_tz("GMT")
    <StaticTzInfo 'GMT'>
    >>> parse_tz("Europe/Paris")
    <DstTzInfo 'Europe/Paris' LMT+0:09:00 STD>
    >>> parse_tz("GMT+1")
    pytz.FixedOffset(60)
    >>> parse_tz("GMT+1:30")
    pytz.FixedOffset(90)
    >>> parse_tz("MalformedString")  # returns None
    """
    try:
        return pytz.timezone(tz)
    except UnknownTimeZoneError:
        GMT_PATTERN = r'GMT(?P<sign>[+-])(?P<hours>[0-9]{1,2})(?::(?P<minutes>[0-9]{2}))?'
        match = re.match(GMT_PATTERN, tz)
        if match:
            return parse_gmt_based_offset(match)
        return None


def parse_gmt_based_offset(match):
    # GMT+2 or GMT+2:30 case
    sign, hours, minutes = match.groups()
    sign = -1 if sign == "-" else 1
    try:
        hours = int(hours)
        minutes = int(minutes) if minutes else 0
    except ValueError:
        return None

    if 0 <= hours < 24 and 0 <= minutes < 60:
        offset = sign * (hours * 60 + minutes)
        return pytz.FixedOffset(offset)
    return None


def half_up_round(value, scale):
    """
    Round values using the "half up" logic

    See more: https://docs.python.org/3/library/decimal.html#rounding-modes

    >>> half_up_round(7.5, 0)
    8.0
    >>> half_up_round(6.5, 0)
    7.0
    >>> half_up_round(-7.5, 0)
    -8.0
    >>> half_up_round(-6.5, 0)
    -7.0
    """
    # Python2 and Python3's round behavior differs for rounding e.g. 0.5
    # hence we handle the "half" case so that it is rounded up
    scaled_value = (value * (10 ** scale))
    removed_part = scaled_value % 1
    if removed_part == 0.5:
        sign = -1 if value < 0 else 1
        value += 10 ** -(scale + 1) * sign
    return round(value, scale)


def half_even_round(value, scale):
    """
    Round values using the "half even" logic

    See more: https://docs.python.org/3/library/decimal.html#rounding-modes

    >>> half_even_round(7.5, 0)
    8.0
    >>> half_even_round(6.5, 0)
    6.0
    >>> half_even_round(-7.5, 0)
    -8.0
    >>> half_even_round(-6.5, 0)
    -6.0
    """
    # Python2 and Python3's round behavior differs for rounding e.g. 0.5
    # hence we handle the "half" case so that it round even up and odd down
    if scale > 0:
        return round(value, scale)
    scaled_value = (value * (10 ** scale))
    removed_part = scaled_value % 1
    if removed_part == 0.5:
        rounded_part = int(scaled_value)
        is_even = (rounded_part + max(0, scale)) % 2 == 0
        sign = -1 if value < 0 else 1
        if is_even:
            value -= 10 ** -(scale + 1) * sign
        else:
            value += 10 ** -(scale + 1) * sign
    return round(value, scale)


def levenshtein_distance(str1, str2):
    if str1 == "":
        return len(str2)
    if str2 == "":
        return len(str1)
    return min(
        levenshtein_distance(str1[1:], str2[1:]) + (str1[0] != str2[0]),
        levenshtein_distance(str1[1:], str2) + 1,
        levenshtein_distance(str1, str2[1:]) + 1
    )


class NotSupportedByThisSparkVersion(Exception):
    pass


def _decorate_with(decorator):
    def inner(obj):
        if inspect.isclass(obj):
            # Process it as a class, decorating all methods of the class
            for name, method in inspect.getmembers(obj, inspect.isroutine):
                setattr(obj, name, decorator(method))
            return obj

        # Otherwise, just assume it's a method here
        return decorator(obj)

    return inner


def since(need_spark_version: str):
    """
    Decorator that tells gelanis to only allow access to this method when the spark version is higher or equal to
      the need_spark_version.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if config.spark_version is not None and config.spark_version < need_spark_version:
                raise NotSupportedByThisSparkVersion

            return func(*args, **kwargs)
        return wrapper

    return _decorate_with(decorator)


def until(gone_at_spark_version: str):
    """
    Decorator that tells gelanis to only allow access to this method when the spark version is lower than the
      gone_at_spark_version.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if config.spark_version is not None and config.spark_version >= gone_at_spark_version:
                raise NotSupportedByThisSparkVersion

            return func(*args, **kwargs)

        return wrapper

    return _decorate_with(decorator)



from io import BytesIO, StringIO
import logging

from ...exceptions import ConnectionException, FileSystemNotSupported
from .file_system import FileSystem

log = logging.getLogger(__name__)

try:
    import requests
except ImportError:
    requests = None


class Http(FileSystem):
    """:class:`.FileSystem` implementation for HTTP."""

    def __init__(self, file_name):
        if requests is None:
            raise FileSystemNotSupported(
                'http not supported. Install "requests".'
            )

        super().__init__(file_name)
        self.headers = None

    @staticmethod
    def resolve_filenames(expr):
        if Http(expr).exists():
            return [expr]
        return []

    def exists(self):
        r = requests.head(self.file_name, allow_redirects=True)
        return r.status_code == 200

    def load(self):
        log.debug('Http GET request for %s.', self.file_name)
        r = requests.get(self.file_name, headers=self.headers)
        if r.status_code != 200:
            raise ConnectionException()
        return BytesIO(r.content)

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        # warning: encoding and encoding_errors are ignored
        log.debug('Http GET request for %s.', self.file_name)
        r = requests.get(self.file_name, headers=self.headers)
        if r.status_code != 200:
            raise ConnectionException()
        return StringIO(r.text)

    def dump(self, stream):
        log.debug('Dump to %s with http PUT.', self.file_name)
        requests.put(self.file_name, data=b''.join(stream))
        return self



from ...types import LongType
from .aggregations import Aggregation


class CollectList(Aggregation):
    pretty_name = "collect_list"

    def __init__(self, column):
        super().__init__(column)
        self.column = column
        self.items = []

    def merge(self, row, schema):
        self.items.append(self.column.eval(row, schema))

    def mergeStats(self, other, schema):
        self.items += other.items

    def eval(self, row, schema):
        return self.items

    def args(self):
        return (self.column,)


class CollectSet(Aggregation):
    pretty_name = "collect_set"

    def __init__(self, column):
        super().__init__(column)
        self.column = column
        self.items = set()

    def merge(self, row, schema):
        self.items.add(self.column.eval(row, schema))

    def mergeStats(self, other, schema):
        self.items |= other.items

    def eval(self, row, schema):
        return list(self.items)

    def args(self):
        return (self.column,)


class SumDistinct(Aggregation):
    pretty_name = "sum_distinct"

    def __init__(self, column):
        super().__init__(column)
        self.column = column
        self.items = set()

    def merge(self, row, schema):
        self.items.add(self.column.eval(row, schema))

    def mergeStats(self, other, schema):
        self.items |= other.items

    def eval(self, row, schema):
        return sum(self.items)

    def args(self):
        return (self.column,)


class First(Aggregation):
    pretty_name = "first"
    _sentinel = object()

    def __init__(self, column, ignore_nulls):
        super().__init__(column)
        self.column = column
        self.value = self._sentinel
        self.ignore_nulls = ignore_nulls.get_literal_value()

    def merge(self, row, schema):
        if self.value is First._sentinel or (self.ignore_nulls and self.value is None):
            self.value = self.column.eval(row, schema)

    def mergeStats(self, other, schema):
        if self.value is First._sentinel or (self.ignore_nulls and self.value is None):
            self.value = other.value

    def eval(self, row, schema):
        return self.value if self.value is not First._sentinel else None

    def args(self):
        return (
            self.column,
            str(self.ignore_nulls).lower()
        )


class Last(Aggregation):
    pretty_name = "last"
    _sentinel = object()

    def __init__(self, column, ignore_nulls):
        super().__init__(column)
        self.column = column
        self.value = None
        self.ignore_nulls = ignore_nulls.get_literal_value()

    def merge(self, row, schema):
        new_value = self.column.eval(row, schema)
        if not (self.ignore_nulls and new_value is None):
            self.value = new_value

    def mergeStats(self, other, schema):
        if not (self.ignore_nulls and other.value is None):
            self.value = other.value

    def eval(self, row, schema):
        return self.value

    def args(self):
        return (
            self.column,
            str(self.ignore_nulls).lower()
        )


class CountDistinct(Aggregation):
    pretty_name = "count"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns
        self.items = set()

    def merge(self, row, schema):
        self.items.add(tuple(
            col.eval(row, schema) for col in self.columns
        ))

    def mergeStats(self, other, schema):
        self.items += other.items

    def eval(self, row, schema):
        return len(self.items)

    def args(self):
        return f"DISTINCT {','.join(self.columns)}"

    def data_type(self, schema):
        return LongType()


class ApproxCountDistinct(Aggregation):
    pretty_name = "approx_count_distinct"

    def __init__(self, column):
        super().__init__(column)
        self.column = column
        self.items = set()

    def merge(self, row, schema):
        self.items.add(self.column.eval(row, schema))

    def mergeStats(self, other, schema):
        self.items += other.items

    def eval(self, row, schema):
        return len(self.items)

    def args(self):
        return (self.column,)

    def data_type(self, schema):
        return LongType()


__all__ = [
    "SumDistinct", "ApproxCountDistinct", "CollectList", "CollectSet",
    "First", "CountDistinct", "Last"
]



from . import Expression


class SortOrder(Expression):
    sort_order = None

    def __init__(self, column):
        super().__init__(column)
        self.column = column

    def eval(self, row, schema):
        return self.column.eval(row, schema)

    def __str__(self):
        return f"{self.column} {self.sort_order}"

    def args(self):
        return (self.column,)


class AscNullsFirst(SortOrder):
    sort_order = "ASC NULLS FIRST"


class AscNullsLast(SortOrder):
    sort_order = "ASC NULLS LAST"


class DescNullsFirst(SortOrder):
    sort_order = "DESCNULLS FIRST"


class DescNullsLast(SortOrder):
    sort_order = "DESC NULLS LAST"


Asc = AscNullsFirst
Desc = DescNullsLast



import datetime
from functools import lru_cache
import os
import re
import time

from .pandas.utils import require_minimum_pandas_version
from .types import StringType

NO_TIMESTAMP_CONVERSION = object()

JAVA_TIME_FORMAT_TOKENS = re.compile("(([a-zA-Z])\\2*|[^a-zA-Z]+)")

FORMAT_MAPPING = {
    "EEEE": "%A",
    "EEE": "%a",
    "EE": "%a",
    "E": "%a",
    "e": "%w",
    "dd": "%d",
    "d": "%-d",
    "MMMM": "%B",
    "MMM": "%b",
    "MM": "%m",
    "M": "%-m",
    "yyyy": "%Y",
    "yyy": "%Y",
    "yy": "%y",
    "y": "%Y",
    "HH": "%H",
    "H": "%-H",
    "hh": "%I",
    "h": "%-I",
    "a": "%p",
    "mm": "%M",
    "m": "%-M",
    "ss": "%S",
    "s": "%-S",
    "S": "%f",
    "xxxx": "%z",
    "xx": "%z",
    "ZZZ": "%z",
    "ZZ": "%z",
    "Z": "%z",
    "DDD": "%j",
    "D": "%-j",
}


def get_sub_formatter(group):
    token, _ = group

    if token in FORMAT_MAPPING:
        return lambda value: value.strftime(FORMAT_MAPPING[token])

    if token in ("'", "[", "]"):
        return lambda value: ""

    if "S" in token:
        number_of_digits = len(token)
        return lambda value: value.strftime("%f")[:number_of_digits]

    if token == "XXX":
        def timezone_formatter(value):
            tz = value.strftime("%z")
            return f"{tz[:3]}:{tz[3:]}" if tz else ""

        return timezone_formatter

    return lambda value: token


@lru_cache(64)
def get_time_formatter(java_time_format):
    """
    Convert a Java time format to a Python time format.

    This function currently only support a small subset of Java time formats.
    """
    sub_formatters = [
        get_sub_formatter(token)
        for token in JAVA_TIME_FORMAT_TOKENS.findall(java_time_format)
    ]

    def time_formatter(value):
        return "".join(sub_formatter(value) for sub_formatter in sub_formatters)

    return time_formatter


def get_unix_timestamp_parser(java_time_format):
    datetime_parser = get_datetime_parser(java_time_format)

    def time_parser(value):
        dt = datetime_parser(value)
        return int(time.mktime(dt.timetuple()))

    return time_parser


@lru_cache(64)
def get_datetime_parser(java_time_format):
    if java_time_format is None:
        # pylint: disable=import-outside-toplevel, cyclic-import
        from ._casts import cast_to_timestamp

        return lambda value: cast_to_timestamp(value, StringType(), {})

    if java_time_format is NO_TIMESTAMP_CONVERSION:
        return lambda value: None

    python_pattern = ""
    for token, _ in JAVA_TIME_FORMAT_TOKENS.findall(java_time_format):
        python_pattern += FORMAT_MAPPING.get(token, token)
    return lambda value: datetime.datetime.strptime(value, python_pattern)


def _get_local_timezone():
    """ Get local timezone using pytz with environment variable, or dateutil.

    If there is a 'TZ' environment variable, pass it to pandas to use pytz and use it as timezone
    string, otherwise use the special word 'dateutil/:' which means that pandas uses dateutil and
    it reads system configuration to know the system local timezone.

    See also:
    - https://github.com/pandas-dev/pandas/blob/0.19.x/pandas/tslib.pyx#L1753
    - https://github.com/dateutil/dateutil/blob/2.6.1/dateutil/tz/tz.py#L1338
    """
    return os.environ.get('TZ', 'dateutil/:')


def _check_series_localize_timestamps(s, timezone):
    """
    Convert timezone aware timestamps to timezone-naive in the specified timezone or local timezone.

    If the input series is not a timestamp series, then the same series is returned. If the input
    series is a timestamp series, then a converted series is returned.

    :param s: pandas.Series
    :param timezone: the timezone to convert. if None then use local timezone
    :return pandas.Series that have been converted to tz-naive
    """
    require_minimum_pandas_version()

    try:
        # pandas is an optional dependency
        # pylint: disable=import-outside-toplevel
        from pandas.api.types import is_datetime64tz_dtype
    except ImportError as e:
        raise Exception("require_minimum_pandas_version() was not called") from e
    tz = timezone or _get_local_timezone()
    # pylint: disable=W0511
    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
    if is_datetime64tz_dtype(s.dtype):
        return s.dt.tz_convert(tz).dt.tz_localize(None)
    return s


def _check_dataframe_localize_timestamps(pdf, timezone):
    """
    Convert timezone aware timestamps to timezone-naive in the specified timezone or local timezone

    :param pdf: pandas.DataFrame
    :param timezone: the timezone to convert. if None then use local timezone
    :return pandas.DataFrame where any timezone aware columns have been converted to tz-naive
    """
    require_minimum_pandas_version()

    for column, series in pdf.iteritems():
        pdf[column] = _check_series_localize_timestamps(series, timezone)
    return pdf


def _check_series_convert_timestamps_internal(s, timezone):
    """
    Convert a tz-naive timestamp in the specified timezone or local timezone to UTC normalized for
    Spark internal storage

    :param s: a pandas.Series
    :param timezone: the timezone to convert. if None then use local timezone
    :return pandas.Series where if it is a timestamp, has been UTC normalized without a time zone
    """
    require_minimum_pandas_version()

    try:
        # pandas is an optional dependency
        # pylint: disable=import-outside-toplevel
        from pandas.api.types import is_datetime64_dtype, is_datetime64tz_dtype
    except ImportError as e:
        raise Exception("require_minimum_pandas_version() was not called") from e

    # pylint: disable=W0511
    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
    if is_datetime64_dtype(s.dtype):
        # When tz_localize a tz-naive timestamp, the result is ambiguous if the tz-naive
        # timestamp is during the hour when the clock is adjusted backward during due to
        # daylight saving time (dst).
        # E.g., for America/New_York, the clock is adjusted backward on 2015-11-01 2:00 to
        # 2015-11-01 1:00 from dst-time to standard time, and therefore, when tz_localize
        # a tz-naive timestamp 2015-11-01 1:30 with America/New_York timezone, it can be either
        # dst time (2015-01-01 1:30-0400) or standard time (2015-11-01 1:30-0500).
        #
        # Here we explicit choose to use standard time. This matches the default behavior of
        # pytz.
        #
        # Here are some code to help understand this behavior:
        # >>> import datetime
        # >>> import pandas as pd
        # >>> import pytz
        # >>>
        # >>> t = datetime.datetime(2015, 11, 1, 1, 30)
        # >>> ts = pd.Series([t])
        # >>> tz = pytz.timezone('America/New_York')
        # >>>
        # >>> ts.dt.tz_localize(tz, ambiguous=True)
        # 0   2015-11-01 01:30:00-04:00
        # dtype: datetime64[ns, America/New_York]
        # >>>
        # >>> ts.dt.tz_localize(tz, ambiguous=False)
        # 0   2015-11-01 01:30:00-05:00
        # dtype: datetime64[ns, America/New_York]
        # >>>
        # >>> str(tz.localize(t))
        # '2015-11-01 01:30:00-05:00'
        tz = timezone or _get_local_timezone()
        return s.dt.tz_localize(tz, ambiguous=False).dt.tz_convert('UTC')
    if is_datetime64tz_dtype(s.dtype):
        return s.dt.tz_convert('UTC')
    return s


def _check_series_convert_timestamps_localize(s, from_timezone, to_timezone):
    """
    Convert timestamp to timezone-naive in the specified timezone or local timezone

    :param s: a pandas.Series
    :param from_timezone: the timezone to convert from. if None then use local timezone
    :param to_timezone: the timezone to convert to. if None then use local timezone
    :return pandas.Series where if it is a timestamp, has been converted to tz-naive
    """
    require_minimum_pandas_version()

    try:
        # pandas is an optional dependency
        # pylint: disable=import-outside-toplevel
        import pandas as pd
        from pandas.api.types import is_datetime64_dtype, is_datetime64tz_dtype
    except ImportError as e:
        raise Exception("require_minimum_pandas_version() was not called") from e

    from_tz = from_timezone or _get_local_timezone()
    to_tz = to_timezone or _get_local_timezone()
    # pylint: disable=W0511
    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
    if is_datetime64tz_dtype(s.dtype):
        return s.dt.tz_convert(to_tz).dt.tz_localize(None)
    if is_datetime64_dtype(s.dtype) and from_tz != to_tz:
        # `s.dt.tz_localize('tzlocal()')` doesn't work properly when including NaT.
        return s.apply(
            lambda ts: ts.tz_localize(from_tz, ambiguous=False).tz_convert(to_tz).tz_localize(None)
            if ts is not pd.NaT else pd.NaT)
    return s


def _check_series_convert_timestamps_local_tz(s, timezone):
    """
    Convert timestamp to timezone-naive in the specified timezone or local timezone

    :param s: a pandas.Series
    :param timezone: the timezone to convert to. if None then use local timezone
    :return pandas.Series where if it is a timestamp, has been converted to tz-naive
    """
    return _check_series_convert_timestamps_localize(s, None, timezone)


def _check_series_convert_timestamps_tz_local(s, timezone):
    """
    Convert timestamp to timezone-naive in the specified timezone or local timezone

    :param s: a pandas.Series
    :param timezone: the timezone to convert from. if None then use local timezone
    :return pandas.Series where if it is a timestamp, has been converted to tz-naive
    """
    return _check_series_convert_timestamps_localize(s, timezone, None)



from . import Expression


class UserDefinedFunction(Expression):
    def __init__(self, f, return_type, *exprs):
        super().__init__()
        self.f = f
        self.return_type = return_type
        self.exprs = exprs

    def eval(self, row, schema):
        return self.f(*(expr.eval(row, schema) for expr in self.exprs))

    def __str__(self):
        arguments = ', '.join(str(arg) for arg in self.args())
        return f"{self.f.__name__}({arguments})"

    def args(self):
        return self.exprs


__all__ = ["UserDefinedFunction"]



from fnmatch import fnmatch
from io import BytesIO, StringIO
import logging

from ...exceptions import FileSystemNotSupported
from ...utils import parse_file_uri, Tokenizer
from .file_system import FileSystem

log = logging.getLogger(__name__)

try:
    import boto
except ImportError:
    boto = None


class S3(FileSystem):
    """:class:`.FileSystem` implementation for S3.

    Use environment variables ``AWS_SECRET_ACCESS_KEY`` and
    ``AWS_ACCESS_KEY_ID`` for auth and use file paths of the form
    ``s3://bucket_name/filename.txt``.
    """

    #: Keyword arguments for new connections.
    #: Example: set to `{'anon': True}` for anonymous connections.
    connection_kwargs = {}

    _conn = None

    def __init__(self, file_name):
        if boto is None:
            raise FileSystemNotSupported('S3 not supported. Install "boto".')

        super().__init__(file_name)

        # obtain key
        t = Tokenizer(self.file_name)
        t.get_next('://')  # skip scheme
        bucket_name = t.get_next('/')
        key_name = t.get_next()
        conn = self._get_conn()
        bucket = conn.get_bucket(bucket_name, validate=False)
        self.key = bucket.get_key(key_name)
        if not self.key:
            self.key = bucket.new_key(key_name)

    @classmethod
    def _get_conn(cls):
        if not cls._conn:
            if boto is None:
                raise FileSystemNotSupported('S3 not supported. Install "boto".')
            cls._conn = boto.connect_s3(**cls.connection_kwargs)
        return cls._conn

    @classmethod
    def resolve_filenames(cls, expr):
        files = []

        t = Tokenizer(expr)
        scheme = t.get_next('://')
        bucket_name = t.get_next('/')
        prefix = t.get_next(['*', '?'])

        bucket = cls._get_conn().get_bucket(
            bucket_name,
            validate=False
        )
        expr = expr[len(scheme) + 3 + len(bucket_name) + 1:]
        for k in bucket.list(prefix=prefix):
            if fnmatch(k.name, expr) or fnmatch(k.name, expr + '/part*'):
                files.append(f'{scheme}://{bucket_name}/{k.name}')
        return files

    @classmethod
    def resolve_content(cls, expr):
        scheme, bucket_name, folder_path, pattern = parse_file_uri(expr)

        folder_path = folder_path[1:]  # Remove leading slash

        expr = f"{folder_path}{pattern}"
        # Match all files inside folders that match expr
        pattern_expr = f"{expr}{'' if expr.endswith('/') else '/'}*"

        bucket = cls._get_conn().get_bucket(
            bucket_name,
            validate=False
        )
        files = []
        for k in bucket.list(prefix=folder_path):
            if fnmatch(k.name, expr) or fnmatch(k.name, pattern_expr):
                files.append(f'{scheme}://{bucket_name}/{k.name}')
        return files

    def exists(self):
        t = Tokenizer(self.file_name)
        t.get_next('//')  # skip scheme
        bucket_name = t.get_next('/')
        key_name = t.get_next()
        conn = self._get_conn()
        bucket = conn.get_bucket(bucket_name, validate=False)
        return (bucket.get_key(key_name)
                or bucket.list(prefix=f'{key_name}/'))

    def load(self):
        log.debug('Loading %s with size %s.', self.key.name, self.key.size)
        return BytesIO(self.key.get_contents_as_string())

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        log.debug('Loading %s with size %s.', self.key.name, self.key.size)
        return StringIO(
            self.key.get_contents_as_string().decode(encoding, encoding_errors)
        )

    def dump(self, stream):
        log.debug('Dumping to %s.', self.key.name)
        self.key.set_contents_from_file(stream)
        return self

    def make_public(self, recursive=False):
        self.key.make_public(recursive)
        return self



# A large part of this module is extracted from its PySpark counterpart at
# https://spark.apache.org/docs/1.5.0/api/python/_modules/pyspark/accumulators.html
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
>>> from gelanis import Context
>>> sc = Context()
>>> a = sc.accumulator(1)
>>> a.value
1
>>> a.value = 2
>>> a.value
2
>>> a += 5
>>> a.value
7

>>> sc.accumulator(1.0).value
1.0

>>> sc.accumulator(1j).value
1j

>>> rdd = sc.parallelize([1,2,3])
>>> def f(x):
...     global a
...     a += x
>>> rdd.foreach(f)
>>> a.value
13

>>> b = sc.accumulator(0)
>>> def g(x):
...     b.add(x)
>>> rdd.foreach(g)
>>> b.value
6

>>> from gelanis import AccumulatorParam
>>> class VectorAccumulatorParam(AccumulatorParam):
...     def zero(self, value):
...         return [0.0] * len(value)
...     def addInPlace(self, val1, val2):
...         for i in range(len(val1)):
...              val1[i] += val2[i]
...         return val1
>>> va = sc.accumulator([1.0, 2.0, 3.0], VectorAccumulatorParam())
>>> va.value
[1.0, 2.0, 3.0]
>>> def g(x):
...     global va
...     va += [x] * 3
>>> rdd.foreach(g)
>>> va.value
[7.0, 8.0, 9.0]

>>> sc.accumulator([1.0, 2.0, 3.0]) # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
...
TypeError: No default accumulator param for type <type 'list'>
"""


__all__ = ['Accumulator', 'AccumulatorParam']


class Accumulator:
    """
    A shared variable that can be accumulated, i.e., has a commutative and associative "add"
    operation. Tasks can add values to an Accumulator with the ``+=`` operator

    The API supports accumulators for primitive data types like ``int`` and
    ``float``, users can also define accumulators for custom types by providing a custom
    ``AccumulatorParam`` object. Refer to the doctest of this module for an example.
    """

    def __init__(self, value, accum_param):
        """Create a new Accumulator with a given initial value and AccumulatorParam object"""
        self.accum_param = accum_param
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def add(self, term):
        """Adds a term to this accumulator's value"""
        self._value = self.accum_param.addInPlace(self._value, term)

    def __iadd__(self, term):
        """The += operator; adds a term to this accumulator's value"""
        self.add(term)
        return self

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return f"Accumulator<value={self._value}>"


class AccumulatorParam:
    """
    Helper object that defines how to accumulate values of a given type.
    """
    def zero(self, value):
        """
        Provide a "zero value" for the type, compatible in dimensions with the
        provided ``value`` (e.g., a zero vector)
        """
        raise NotImplementedError

    def addInPlace(self, value1, value2):
        """
        Add two values of the accumulator's data type, returning a new value;
        for efficiency, can also update ``value1`` in place and return it.
        """
        raise NotImplementedError


class AddingAccumulatorParam(AccumulatorParam):
    """
    An AccumulatorParam that uses the + operators to add values. Designed for simple types
    such as integers, floats, and lists. Requires the zero value for the underlying type
    as a parameter.
    """
    def __init__(self, zero_value):
        self.zero_value = zero_value

    def zero(self, value):
        return self.zero_value

    def addInPlace(self, value1, value2):
        value1 += value2
        return value1


# Singleton accumulator params for some standard types
INT_ACCUMULATOR_PARAM = AddingAccumulatorParam(0)
FLOAT_ACCUMULATOR_PARAM = AddingAccumulatorParam(0.0)
COMPLEX_ACCUMULATOR_PARAM = AddingAccumulatorParam(0.0j)


if __name__ == "__main__":
    #
    # Execute doctests with
    #
    # $ python -m gelanis.accumulators -v
    #
    import doctest
    import sys

    failure_count, _ = doctest.testmod()
    if failure_count:
        sys.exit(-1)



import logging

log = logging.getLogger(__name__)


class FileSystem:
    """Interface class for the file system.

    :param str file_name: File name.
    """
    def __init__(self, file_name):
        self.file_name = file_name

    @staticmethod
    def resolve_filenames(expr):
        """Resolve the given glob-like expression to filenames.

        :rtype: list
        """
        log.error('Cannot resolve: %s', expr)

    @staticmethod
    def resolve_content(expr):
        """Return all the files matching expr or in a folder matching expr

        :rtype: list
        """
        log.error('Cannot resolve: %s', expr)

    def exists(self):
        """Check whether the given file_name exists.

        :rtype: bool
        """
        log.warning('Could not determine whether %s exists due to unhandled scheme.', self.file_name)

    def load(self):
        """Load a file to a stream.

        :rtype: io.BytesIO
        """
        log.error('Cannot load: %s', self.file_name)

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        """Load a file to a stream.

        :param str encoding: Text encoding.
        :param str encoding_errors: How to handle encoding errors.

        :rtype: io.StringIO
        """
        log.error('Cannot load: %s', self.file_name)

    def dump(self, stream):
        """Dump a stream to a file.

        :param io.BytesIO stream: Input tream.
        """
        log.error('Cannot dump: %s', self.file_name)

    def make_public(self, recursive=False):
        """Make the file public (only on some file systems).

        :param bool recursive: Recurse.
        :rtype: FileSystem
        """
        log.warning('Cannot make %s public.', self.file_name)



import logging

from .bz2 import Bz2
from .codec import Codec
from .gz import Gz
from .lzma import Lzma
from .sevenz import SevenZ
from .tar import Tar, TarBz2, TarGz
from .zip import Zip

log = logging.getLogger(__name__)

FILE_ENDINGS = [
    (('.tar',), Tar),
    (('.tar.gz',), TarGz),
    (('.tar.bz2',), TarBz2),
    (('.gz',), Gz),
    (('.zip',), Zip),
    (('.bz2',), Bz2),
    (('.lzma', '.xz'), Lzma),
    (('.7z',), SevenZ),
]


class NoCodec(Codec):
    pass


def get_codec(path):
    """Find the codec implementation for this path."""
    if '.' not in path or path.rfind('/') > path.rfind('.'):
        return Codec

    for endings, codec_class in FILE_ENDINGS:
        if any(path.endswith(e) for e in endings):
            log.debug('Using %s codec: %s', endings, path)
            return codec_class

    return NoCodec



from ..._statcounter import CovarianceCounter
from .aggregations import Aggregation


class CovarianceStatAggregation(Aggregation):
    def __init__(self, column1, column2):
        super().__init__(column1, column2)
        self.column1 = column1
        self.column2 = column2
        self.stat_helper = CovarianceCounter(method="pearson")

    def merge(self, row, schema):
        self.stat_helper.add(row.eval(self.column1, schema), row.eval(self.column2, schema))

    def mergeStats(self, other, schema):
        self.stat_helper.merge(other)

    def eval(self, row, schema):
        raise NotImplementedError

    def args(self):
        return (
            self.column1,
            self.column2
        )


class Corr(CovarianceStatAggregation):
    pretty_name = "corr"

    def eval(self, row, schema):
        return self.stat_helper.pearson_correlation


class CovarSamp(CovarianceStatAggregation):
    pretty_name = "covar_samp"

    def eval(self, row, schema):
        return self.stat_helper.covar_samp


class CovarPop(CovarianceStatAggregation):
    pretty_name = "covar_pop"

    def eval(self, row, schema):
        return self.stat_helper.covar_pop


__all__ = ["Corr", "CovarSamp", "CovarPop"]



from setuptools import find_packages, setup
import versioneer

setup(
    name='gelanis',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    license='MIT',
    description='Pure Python implementation of the Spark RDD interface.',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    author='pysparkling contributors',
    url='https://github.com/kbc-opensource/pysparkling',

    install_requires=[
        'boto>=2.36.0',
        'future>=0.15',
        'requests>=2.6.0',
        'pytz>=2019.3',
        'python-dateutil>=2.8.0',
        'pythonsqlparser',
    ],
    extras_require={
        'hdfs': ['hdfs>=2.0.0'],
        'performance': ['matplotlib>=1.5.3'],
        'streaming': ['tornado>=4.3'],
        'dev': [
            'antlr4-python3-runtime==4.7.1',
        ],
        'sql': [
            'numpy',
            'pandas>=0.23.2',
        ],
        'tests': [
            'backports.tempfile==1.0rc1',
            'cloudpickle>=0.1.0',
            'futures>=3.0.1',
            'pylint~=2.7',
            'pylzma',
            'memory-profiler>=0.47',
            'pycodestyle',
            'pytest',
            'pytest-cov',
            'isort',
            'tornado>=4.3',
        ],
        'scripts': [
            'ipyparallel',
            'pyspark',
            'matplotlib',
        ]
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: PyPy',
    ]
)



import base64
import math
import random
import re
import string

from ...utils import half_even_round, half_up_round, MonotonicallyIncreasingIDGenerator, XORShiftRandom
from ..internal_utils.column import resolve_column
from ..types import create_row, StringType
from ..utils import AnalysisException
from .expressions import Expression, NullSafeColumnOperation, UnaryExpression
from .operators import Cast

JVM_MAX_INTEGER_SIZE = 2 ** 63


class StarOperator(Expression):
    @property
    def may_output_multiple_cols(self):
        return True

    def output_fields(self, schema):
        return schema.fields

    def eval(self, row, schema):
        return [row[col] for col in row.__fields__]

    def __str__(self):
        return "*"

    def args(self):
        return ()


class CaseWhen(Expression):
    def __init__(self, conditions, values):
        super().__init__(conditions, values)
        self.conditions = conditions
        self.values = values

    def eval(self, row, schema):
        for condition, function in zip(self.conditions, self.values):
            condition_value = condition.eval(row, schema)
            if condition_value:
                return function.eval(row, schema)
        return None

    def __str__(self):
        when_statements = " ".join(
            f"WHEN {condition} THEN {value}"
            for condition, value in zip(self.conditions, self.values)
        )

        return f"CASE {when_statements} END"

    def args(self):
        return (
            self.conditions,
            self.values
        )

    def add_when(self, condition, value):
        return CaseWhen(
            self.conditions + [condition],
            self.values + [value]
        )

    def set_otherwise(self, default):
        return Otherwise(
            self.conditions,
            self.values,
            default
        )


class Otherwise(Expression):
    def __init__(self, conditions, values, default):
        super().__init__(conditions, values, default)

        self.conditions = conditions
        self.values = values
        self.default = default

    def eval(self, row, schema):
        for condition, function in zip(self.conditions, self.values):
            condition_value = condition.eval(row, schema)
            if condition_value:
                return function.eval(row, schema)
        if self.default is not None:
            return self.default.eval(row, schema)
        return None

    def __str__(self):
        whens = " ".join(
            f"WHEN {condition} THEN {value}"
            for condition, value in zip(self.conditions, self.values)
        )

        return f"CASE {whens} ELSE {self.default} END"

    def args(self):
        return (
            self.conditions,
            self.values,
            self.default
        )


class RegExpExtract(Expression):
    pretty_name = "regexp_extract"

    def __init__(self, e, exp, groupIdx):
        super().__init__(e, exp, groupIdx)

        self.exp = exp.get_literal_value()
        self.groupIdx = groupIdx.get_literal_value()
        self.e = e

        regexp = re.compile(self.exp)

        def fn(x):
            match = regexp.search(x)
            if not match:
                return ""
            ret = match.group(self.groupIdx)
            return ret

        self.fn = fn

    def eval(self, row, schema):
        return self.fn(self.e.eval(row, schema))

    def args(self):
        return (
            self.e,
            self.exp,
            self.groupIdx
        )


class RegExpReplace(Expression):
    pretty_name = "regexp_replace"

    def __init__(self, e, exp, replacement):
        super().__init__(e, exp, replacement)

        self.exp = exp.get_literal_value()
        self.replacement = replacement.get_literal_value()
        self.e = e

        regexp = re.compile(self.exp)
        self.fn = lambda x: regexp.sub(self.replacement, x)

    def eval(self, row, schema):
        return self.fn(self.e.eval(row, schema))

    def args(self):
        return (
            self.e,
            self.exp,
            self.replacement
        )


class Round(NullSafeColumnOperation):
    pretty_name = "round"

    def __init__(self, column, scale):
        super().__init__(column)
        self.scale = scale.get_literal_value()

    def unsafe_operation(self, value):
        return half_up_round(value, self.scale)

    def args(self):
        return (
            self.column,
            self.scale
        )


class Bround(NullSafeColumnOperation):
    pretty_name = "bround"

    def __init__(self, column, scale):
        super().__init__(column)
        self.scale = scale.get_literal_value()

    def unsafe_operation(self, value):
        return half_even_round(value, self.scale)

    def args(self):
        return (
            self.column,
            self.scale
        )


class FormatNumber(Expression):
    pretty_name = "format_number"

    def __init__(self, column, digits):
        super().__init__(column)
        self.column = column
        self.digits = digits.get_literal_value()

    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        if self.digits < 0:
            return None
        if not isinstance(value, (int, float)):
            return None
        rounded_value = half_even_round(value, self.digits)
        return f"{rounded_value:,}"

    def args(self):
        return (
            self.column,
            self.digits
        )


class SubstringIndex(Expression):
    pretty_name = "substring_index"

    def __init__(self, column, delim, count):
        super().__init__(column)
        self.column = column
        self.delim = delim.get_literal_value()
        self.count = count.get_literal_value()

    def eval(self, row, schema):
        parts = str(self.column.eval(row, schema)).split(self.delim)
        return self.delim.join(parts[:self.count] if self.count > 0 else parts[self.count:])

    def args(self):
        return (
            self.column,
            self.delim,
            self.count
        )


class Coalesce(Expression):
    pretty_name = "coalesce"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns

    def eval(self, row, schema):
        for col in self.columns:
            col_value = col.eval(row, schema)
            if col_value is not None:
                return col_value
        return None

    def args(self):
        return self.columns


class IsNaN(UnaryExpression):
    pretty_name = "isnan"

    def eval(self, row, schema):
        return math.isnan(self.eval(row, schema))


class NaNvl(Expression):
    pretty_name = "nanvl"

    def __init__(self, col1, col2):
        super().__init__(col1, col2)
        self.col1 = col1
        self.col2 = col2

    def eval(self, row, schema):
        nan = float("nan")
        col1_value = self.col1.eval(row, schema)
        if col1_value is not nan:
            return float(col1_value)
        return float(self.col2.eval(row, schema))

    def args(self):
        return (
            self.col1,
            self.col2
        )


class Hypot(Expression):
    pretty_name = "hypot"

    def __init__(self, a, b):
        super().__init__(a, b)
        self.a = a
        self.b = b

    def eval(self, row, schema):
        return math.hypot(self.a, self.b)

    def args(self):
        return (
            self.a,
            self.b
        )


class Sqrt(UnaryExpression):
    pretty_name = "SQRT"

    def eval(self, row, schema):
        return math.sqrt(self.column.eval(row, schema))


class Cbrt(UnaryExpression):
    pretty_name = "CBRT"

    def eval(self, row, schema):
        return self.column.eval(row, schema) ** 1. / 3.


class Abs(UnaryExpression):
    pretty_name = "ABS"

    def eval(self, row, schema):
        return abs(self.column.eval(row, schema))


class Acos(UnaryExpression):
    pretty_name = "ACOS"

    def eval(self, row, schema):
        return math.acos(self.column.eval(row, schema))


class Asin(UnaryExpression):
    pretty_name = "ASIN"

    def eval(self, row, schema):
        return math.asin(self.column.eval(row, schema))


class Atan(UnaryExpression):
    pretty_name = "ATAN"

    def eval(self, row, schema):
        return math.atan(self.column.eval(row, schema))


class Atan2(Expression):
    pretty_name = "ATAN"

    def __init__(self, y, x):
        super().__init__(y, x)
        self.y = y
        self.x = x

    def eval(self, row, schema):
        return math.atan2(self.y.eval(row, schema), self.x.eval(row, schema))

    def args(self):
        return (
            self.y,
            self.x
        )


class Tan(UnaryExpression):
    pretty_name = "TAN"

    def eval(self, row, schema):
        return math.tan(self.column.eval(row, schema))


class Tanh(UnaryExpression):
    pretty_name = "TANH"

    def eval(self, row, schema):
        return math.tanh(self.column.eval(row, schema))


class Cos(UnaryExpression):
    pretty_name = "COS"

    def eval(self, row, schema):
        return math.cos(self.column.eval(row, schema))


class Cosh(UnaryExpression):
    pretty_name = "COSH"

    def eval(self, row, schema):
        return math.cosh(self.column.eval(row, schema))


class Sin(UnaryExpression):
    pretty_name = "SIN"

    def eval(self, row, schema):
        return math.sin(self.column.eval(row, schema))


class Sinh(UnaryExpression):
    pretty_name = "SINH"

    def eval(self, row, schema):
        return math.sinh(self.column.eval(row, schema))


class Exp(UnaryExpression):
    pretty_name = "EXP"

    def eval(self, row, schema):
        return math.exp(self.column.eval(row, schema))


class ExpM1(UnaryExpression):
    pretty_name = "EXPM1"

    def eval(self, row, schema):
        return math.expm1(self.column.eval(row, schema))


class Factorial(UnaryExpression):
    pretty_name = "factorial"

    def eval(self, row, schema):
        return math.factorial(self.column.eval(row, schema))


class Floor(UnaryExpression):
    pretty_name = "FLOOR"

    def eval(self, row, schema):
        return math.floor(self.column.eval(row, schema))


class Ceil(UnaryExpression):
    pretty_name = "CEIL"

    def eval(self, row, schema):
        return math.ceil(self.column.eval(row, schema))


class Log(Expression):
    pretty_name = "LOG"

    def __init__(self, base, value):
        super().__init__(base, value)
        self.base = base.get_literal_value()
        self.value = value

    def eval(self, row, schema):
        value_eval = self.value.eval(row, schema)
        if value_eval == 0:
            return None
        return math.log(value_eval, self.base)

    def args(self):
        if self.base == math.e:
            return (self.value, )
        return (
            self.base,
            self.value
        )


class Log10(UnaryExpression):
    pretty_name = "LOG10"

    def eval(self, row, schema):
        return math.log10(self.column.eval(row, schema))


class Log2(UnaryExpression):
    pretty_name = "LOG2"

    def eval(self, row, schema):
        return math.log(self.column.eval(row, schema), 2)


class Log1p(UnaryExpression):
    pretty_name = "LOG1P"

    def eval(self, row, schema):
        return math.log1p(self.column.eval(row, schema))


class Rint(UnaryExpression):
    pretty_name = "ROUND"

    def eval(self, row, schema):
        return round(self.column.eval(row, schema))


class Signum(UnaryExpression):
    pretty_name = "SIGNUM"

    def eval(self, row, schema):
        column_value = self.column.eval(row, schema)
        if column_value == 0:
            return 0
        if column_value > 0:
            return 1.0
        return -1.0


class ToDegrees(UnaryExpression):
    pretty_name = "DEGREES"

    def eval(self, row, schema):
        return math.degrees(self.column.eval(row, schema))


class ToRadians(UnaryExpression):
    pretty_name = "RADIANS"

    def eval(self, row, schema):
        return math.radians(self.column.eval(row, schema))


class Rand(Expression):
    pretty_name = "rand"

    def __init__(self, seed=None):
        super().__init__()
        self.seed = seed.get_literal_value() if seed is not None else random.random()
        self.random_generator = None

    def eval(self, row, schema):
        return self.random_generator.nextDouble()

    def initialize(self, partition_index):
        self.random_generator = XORShiftRandom(self.seed + partition_index)

    def args(self):
        return self.seed


class Randn(Expression):
    pretty_name = "randn"

    def __init__(self, seed=None):
        super().__init__()
        self.seed = seed.get_literal_value()
        self.random_generator = None

    def eval(self, row, schema):
        return self.random_generator.nextGaussian()

    def initialize(self, partition_index):
        self.random_generator = XORShiftRandom(self.seed + partition_index)

    def args(self):
        return self.seed


class SparkPartitionID(Expression):
    pretty_name = "SPARK_PARTITION_ID"

    def __init__(self):
        super().__init__()
        self.partition_index = None

    def eval(self, row, schema):
        return self.partition_index

    def initialize(self, partition_index):
        self.partition_index = partition_index

    def args(self):
        return ()


class CreateStruct(Expression):
    pretty_name = "struct"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns

    def eval(self, row, schema):
        struct_cols, struct_values = [], []
        for col in self.columns:
            output_cols, output_values = resolve_column(col, row, schema, allow_generator=False)
            struct_cols += output_cols
            struct_values += output_values[0]
        return create_row(struct_cols, struct_values)

    def args(self):
        return self.columns


class Bin(UnaryExpression):
    pretty_name = "bin"

    def eval(self, row, schema):
        return format(self.column.eval(row, schema), 'b')


class ShiftLeft(Expression):
    pretty_name = "shiftleft"

    def __init__(self, arg, num_bits):
        super().__init__(arg)
        self.arg = arg
        self.num_bits = num_bits.get_literal_value()

    def eval(self, row, schema):
        return self.arg.eval(row, schema) << self.num_bits

    def args(self):
        return (
            self.arg,
            self.num_bits
        )


class ShiftRight(Expression):
    pretty_name = "shiftright"

    def __init__(self, arg, num_bits):
        super().__init__(arg)
        self.arg = arg
        self.num_bits = num_bits.get_literal_value()

    def eval(self, row, schema):
        return self.arg.eval(row, schema) >> self.num_bits

    def args(self):
        return (
            self.arg,
            self.num_bits
        )


class ShiftRightUnsigned(Expression):
    pretty_name = "shiftrightunsigned"

    def __init__(self, arg, num_bits):
        super().__init__(arg)
        self.arg = arg
        self.num_bits = num_bits.get_literal_value()

    def eval(self, row, schema):
        rightShifted = self.arg.eval(row, schema) >> self.num_bits
        return rightShifted % JVM_MAX_INTEGER_SIZE

    def args(self):
        return (
            self.arg,
            self.num_bits
        )


class Greatest(Expression):
    pretty_name = "greatest"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns

    def eval(self, row, schema):
        values = (col.eval(row, schema) for col in self.columns)
        return max((value for value in values if value is not None), default=None)

    def args(self):
        return self.columns


class Least(Expression):
    pretty_name = "least"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns

    def eval(self, row, schema):
        values = (col.eval(row, schema) for col in self.columns)
        return min((value for value in values if value is not None), default=None)

    def args(self):
        return self.columns


class Length(UnaryExpression):
    pretty_name = "length"

    def eval(self, row, schema):
        return len(str(self.column.eval(row, schema)))


class Lower(UnaryExpression):
    pretty_name = "lower"

    def eval(self, row, schema):
        return str(self.column.eval(row, schema)).lower()


class Upper(UnaryExpression):
    pretty_name = "upper"

    def eval(self, row, schema):
        return str(self.column.eval(row, schema)).upper()


class Concat(Expression):
    pretty_name = "concat"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns

    def eval(self, row, schema):
        return "".join(str(col.eval(row, schema)) for col in self.columns)

    def args(self):
        return self.columns


class ConcatWs(Expression):
    pretty_name = "concat_ws"

    def __init__(self, sep, *columns):
        super().__init__(columns)
        self.sep = sep.get_literal_value()
        self.columns = columns

    def eval(self, row, schema):
        return self.sep.join(str(col.eval(row, schema)) for col in self.columns)

    def args(self):
        if self.columns:
            return [self.sep] + list(self.columns)
        return [self.sep]


class Reverse(UnaryExpression):
    pretty_name = "reverse"

    def eval(self, row, schema):
        return str(self.column.eval(row, schema))[::-1]


class MapKeys(UnaryExpression):
    pretty_name = "map_keys"

    def eval(self, row, schema):
        return list(self.column.eval(row, schema).keys())


class MapValues(UnaryExpression):
    pretty_name = "map_values"

    def eval(self, row, schema):
        return list(self.column.eval(row, schema).values())


class MapEntries(UnaryExpression):
    pretty_name = "map_entries"

    def eval(self, row, schema):
        return list(self.column.eval(row, schema).items())


class MapFromEntries(UnaryExpression):
    pretty_name = "map_from_entries"

    def eval(self, row, schema):
        return dict(self.column.eval(row, schema))


class MapConcat(Expression):
    pretty_name = "map_concat"

    def __init__(self, *columns):
        super().__init__(*columns)
        self.columns = columns

    def eval(self, row, schema):
        result = {}
        for column in self.columns:
            col_value = column.eval(row, schema)
            if isinstance(col_value, dict):
                result.update(col_value)
        return result

    def args(self):
        return self.columns


class StringSplit(Expression):
    pretty_name = "split"

    def __init__(self, column, regex, limit):
        super().__init__(column)
        self.column = column
        self.regex = regex.get_literal_value()
        self.compiled_regex = re.compile(self.regex)
        self.limit = limit.get_literal_value()

    def eval(self, row, schema):
        limit = self.limit if self.limit is not None else 0
        return list(self.compiled_regex.split(str(self.column.eval(row, schema)), limit))

    def args(self):
        if self.limit is None:
            return (
                self.column,
                self.regex
            )
        return (
            self.column,
            self.regex,
            self.limit
        )


class Conv(Expression):
    pretty_name = "conv"

    def __init__(self, column, from_base, to_base):
        super().__init__(column)
        self.column = column
        self.from_base = from_base.get_literal_value()
        self.to_base = to_base.get_literal_value()

    def eval(self, row, schema):
        value = Cast(self.column, StringType()).eval(row, schema)
        return self.convert(
            value,
            self.from_base,
            abs(self.to_base),
            positive_only=self.to_base > 0
        )

    def args(self):
        return (
            self.column,
            self.from_base,
            self.to_base
        )

    @staticmethod
    def convert(from_string, from_base, to_base, positive_only=False):
        """
        from_string: from number as a string
        from_base: from base
        raw_to_base: to base

        Convert a string representation of a number in base from_base to base raw_to_base

        Both base absolute values must be between 2 and 36
        otherwise the function returns None.

        from_base must be positive
        If to_base is

        >>> Conv.convert("1248", 10, 10)
        '1248'
        >>> Conv.convert("1548", 10, 2)
        '11000001100'
        >>> Conv.convert("44953", 10, 36)
        'YOP'
        >>> Conv.convert("YOP", 36, 10)
        '44953'
        >>> Conv.convert("1234", 5, 10)
        '194'
        >>> Conv.convert("-1", 36, 10)
        '-1'
        >>> Conv.convert("-1", 36, 10, positive_only=True)
        '18446744073709551615'
        >>> Conv.convert("YOP", 1, 10)  # returns None if from_base < 2
        >>> Conv.convert("YOP", 40, 10)  # returns None if from_base > 36
        >>> Conv.convert("YOP", 36, 40)  # returns None if to_base > 36
        >>> Conv.convert("YOP", 36, 0)  # returns None if to_base < 2
        >>> Conv.convert("YOP", 10, 2)  # returns None if value is not in the from_base
        """
        if (not (2 <= from_base <= 36 and 2 <= to_base <= 36)
                or from_string is None
                or not from_string):
            return None

        if from_string.startswith("-"):
            value_is_negative = True
            from_numbers = from_string[1:]
        else:
            value_is_negative = False
            from_numbers = from_string

        digits = string.digits + string.ascii_uppercase
        if not set(digits[:from_base]).issuperset(set(from_numbers)):
            return None

        value = sum(
            digits.index(digit) * (from_base ** i)
            for i, digit in enumerate(from_numbers[::-1])
        )

        if value_is_negative and positive_only:
            value = 2 ** 64 - value

        returned_string = ""
        for exp in range(int(math.log(value, to_base)) + 1, -1, -1):
            factor = (to_base ** exp)
            number = value // factor
            value -= number * factor
            returned_string += digits[number]

        if returned_string:
            returned_string = returned_string.lstrip("0")

        if value_is_negative and not positive_only:
            returned_string = "-" + returned_string

        return returned_string


class Hex(UnaryExpression):
    pretty_name = "hex"

    def eval(self, row, schema):
        return Conv.convert(
            self.column.eval(row, schema),
            from_base=10,
            to_base=16,
            positive_only=True
        )


class Unhex(UnaryExpression):
    pretty_name = "unhex"

    def eval(self, row, schema):
        return Conv.convert(
            self.column.eval(row, schema),
            from_base=16,
            to_base=10,
            positive_only=True
        )


class Ascii(UnaryExpression):
    pretty_name = "ascii"

    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        if value is None:
            return None
        value_as_string = str(value)
        if not value_as_string:
            return None
        return ord(value_as_string[0])


class MonotonicallyIncreasingID(Expression):
    pretty_name = "monotonically_increasing_id"

    def __init__(self):
        super().__init__()
        self.generator = None

    def eval(self, row, schema):
        return next(self.generator)

    def initialize(self, partition_index):
        self.generator = MonotonicallyIncreasingIDGenerator(partition_index)

    def args(self):
        return ()


class Base64(UnaryExpression):
    pretty_name = "base64"

    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        encoded = base64.b64encode(bytes(value, encoding="utf-8"))
        return str(encoded)[2:-1]


class UnBase64(UnaryExpression):
    pretty_name = "unbase64"

    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        return bytearray(base64.b64decode(value))


class GroupingID(Expression):
    pretty_name = "grouping_id"

    def __init__(self, *columns):
        super().__init__(*columns)
        self.columns = columns

    def eval(self, row, schema):
        metadata = row.get_metadata()
        if metadata is None or "grouping" not in metadata:
            raise AnalysisException("grouping_id() can only be used with GroupingSets/Cube/Rollup")
        id_binary_string_value = "".join(
            "1" if grouping else "0" for grouping in metadata["grouping"]
        )
        return int(id_binary_string_value, 2)

    def args(self):
        return self.columns


class Grouping(UnaryExpression):
    pretty_name = "grouping"

    def eval(self, row, schema):
        metadata = row.get_metadata()
        if metadata is None or "grouping" not in metadata:
            raise AnalysisException("grouping_id() can only be used with GroupingSets/Cube/Rollup")
        pos = self.column.find_position_in_schema(schema)
        return int(metadata["grouping"][pos])


class InputFileName(Expression):
    pretty_name = "input_file_name"

    def eval(self, row, schema):
        metadata = row.get_metadata()
        if metadata is None:
            return None
        return metadata.get("input_file_name", "")

    def args(self):
        return ()


__all__ = [
    "Grouping", "GroupingID", "Coalesce", "IsNaN", "MonotonicallyIncreasingID", "NaNvl", "Rand",
    "Randn", "SparkPartitionID", "Sqrt", "CreateStruct", "CaseWhen", "Abs", "Acos", "Asin",
    "Atan", "Atan2", "Bin", "Cbrt", "Ceil", "Conv", "Cos", "Cosh", "Exp", "ExpM1", "Factorial",
    "Floor", "Greatest", "Hex", "Unhex", "Hypot", "Least", "Log", "Log10", "Log1p", "Log2",
    "Rint", "Round", "Bround", "Signum", "Sin", "Sinh", "Tan", "Tanh", "ToDegrees",
    "ToRadians", "Ascii", "Base64", "ConcatWs", "FormatNumber", "Length", "Lower",
    "RegExpExtract", "RegExpReplace", "UnBase64", "StringSplit", "SubstringIndex", "Upper",
    "Concat", "Reverse", "MapKeys", "MapValues", "MapEntries", "MapFromEntries",
    "MapConcat", "StarOperator", "ShiftLeft", "ShiftRight", "ShiftRightUnsigned"
]



from functools import partial
import itertools

from ....fileio import TextFile
from ...internal_utils.options import Options
from ...internal_utils.readers.utils import resolve_partitions
from ...types import create_row, StringType, StructField, StructType


class TextReader:
    default_options = dict(
        lineSep=None,
        encoding="utf-8",
        sep=",",
        inferSchema=False,
        header=False
    )

    def __init__(self, spark, paths, schema, options):
        self.spark = spark
        self.paths = paths
        self.schema = schema or StructType([StructField("value", StringType())])
        self.options = Options(self.default_options, options)

    def read(self):
        sc = self.spark._sc
        paths = self.paths

        partitions, partition_schema = resolve_partitions(paths)

        rdd_filenames = sc.parallelize(sorted(partitions.keys()), len(partitions))
        rdd = rdd_filenames.flatMap(partial(
            parse_text_file,
            partitions,
            partition_schema,
            self.schema,
            self.options
        ))

        if partition_schema:
            partitions_fields = partition_schema.fields
            full_schema = StructType(self.schema.fields + partitions_fields)
        else:
            full_schema = self.schema

        rdd._name = paths

        # pylint: disable=import-outside-toplevel, cyclic-import
        from ...internals import DataFrameInternal

        return DataFrameInternal(
            sc,
            rdd,
            schema=full_schema
        )


def parse_text_file(partitions, partition_schema, schema, options, file_name):
    f_content = TextFile(file_name).load(encoding=options.encoding).read()
    records = (f_content.split(options.lineSep)
               if options.lineSep is not None
               else f_content.splitlines())

    rows = []
    for record in records:
        row = text_record_to_row(record, options, schema, partition_schema, partitions[file_name])
        row.set_input_file_name(file_name)
        rows.append(row)

    return rows


def text_record_to_row(record, options, schema, partition_schema, partition):
    partition_field_names = [
        f.name for f in partition_schema.fields
    ] if partition_schema else []
    row = create_row(
        itertools.chain([schema.fields[0].name], partition_field_names),
        itertools.chain([record], partition or [])
    )
    return row



import json

from . import Expression
from .._json import get_json_encoder
from ..internal_utils.options import Options
from ..internal_utils.readers.jsonreader import JSONReader


class StructsToJson(Expression):
    pretty_name = "structstojson"

    default_options = dict(
        dateFormat="yyyy-MM-dd",
        timestampFormat="yyyy-MM-dd'T'HH:mm:ss.SSSXXX",
    )

    def __init__(self, column, options):
        super().__init__(column)
        self.column = column
        self.input_options = options
        self.options = Options(JSONReader.default_options, options)
        self.encoder = get_json_encoder(self.options)

    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        return json.dumps(
            value,
            cls=self.encoder,
            separators=(',', ':')
        )

    def args(self):
        if self.input_options is None:
            return (self.column, )
        return (
            self.column,
            self.input_options
        )


__all__ = ["StructsToJson"]



from . import functions
from .column import Column

__all__ = ['GroupedData']


class GroupedData:
    def __init__(self, jgd, df):
        self._jgd = jgd
        self._df = df
        self.sql_ctx = df.sql_ctx

    def agg(self, *exprs):
        """
        # >>> sorted(gdf.agg({"*": "count"}).collect())
        # [Row(name='Alice', count(1)=1), Row(name='Bob', count(1)=1)]

        >>> from gelanis import Context, Row
        >>> from gelanis.sql.session import SparkSession
        >>> from gelanis.sql import functions
        >>> spark = SparkSession(Context())
        >>> df = spark.createDataFrame(
        ...   [Row(age=2, name='Alice'), Row(age=5, name='Bob')]
        ... )
        >>> gdf = df.groupBy(df.name)
        >>> from gelanis.sql import functions
        >>> sorted(gdf.agg(functions.min(df.age)).collect())
        [Row(name='Alice', min(age)=2), Row(name='Bob', min(age)=5)]
        >>> df.groupBy("age").agg(functions.avg("age"), functions.col("age")).show()
        +---+--------+---+
        |age|avg(age)|age|
        +---+--------+---+
        |  2|     2.0|  2|
        |  5|     5.0|  5|
        +---+--------+---+

        """
        if not exprs:
            raise ValueError("exprs should not be empty")

        if len(exprs) == 1 and isinstance(exprs[0], dict):
            # pylint: disable=W0511
            # todo implement agg_dict
            jdf = self._jgd.agg_dict(exprs[0])
        else:
            # Columns
            if not all(isinstance(c, Column) for c in exprs):
                raise ValueError("all exprs should be Column")

            # noinspection PyProtectedMember
            jdf = self._jgd.agg([functions.parse(e) for e in exprs])

        # pylint: disable=import-outside-toplevel, cyclic-import
        from .dataframe import DataFrame

        return DataFrame(jdf, self.sql_ctx)

    def count(self):
        return self.agg(functions.count(functions.lit(1)).alias("count"))

    # pylint: disable=W0511
    # todo: avg, max, etc should work when cols is left empty
    def mean(self, *cols):
        return self.agg(*(functions.mean(functions.parse(col)) for col in cols))

    def avg(self, *cols):
        return self.agg(*(functions.avg(functions.parse(col)) for col in cols))

    def max(self, *cols):
        return self.agg(*(functions.max(functions.parse(col)) for col in cols))

    def min(self, *cols):
        return self.agg(*(functions.min(functions.parse(col)) for col in cols))

    def sum(self, *cols):
        return self.agg(*(functions.sum(functions.parse(col)) for col in cols))

    def pivot(self, pivot_col, values=None):
        """
        >>> from gelanis import Context, Row
        >>> from gelanis.sql.session import SparkSession
        >>> from gelanis.sql import functions as functions
        >>> sc = Context()
        >>> spark = SparkSession(sc)
        >>> df4 = sc.parallelize([Row(course="dotNET", year=2012, earnings=10000),
        ...                        Row(course="Java",   year=2012, earnings=20000),
        ...                        Row(course="dotNET", year=2012, earnings=5000),
        ...                        Row(course="dotNET", year=2013, earnings=48000),
        ...                        Row(course="Java",   year=2013, earnings=30000)]).toDF()
        >>> df4.groupBy("year").pivot("course", ["dotNET", "Java"]).sum("earnings").collect()
        [Row(year=2012, dotNET=15000, Java=20000), Row(year=2013, dotNET=48000, Java=30000)]
        >>> df4.groupBy("year").pivot("course").sum("earnings").collect()
        [Row(year=2012, Java=20000, dotNET=15000), Row(year=2013, Java=30000, dotNET=48000)]
        >>> df4.groupBy("year").pivot("course", ["dotNET"]).sum("earnings").collect()
        [Row(year=2012, dotNET=15000), Row(year=2013, dotNET=48000)]
        >>> df4.groupBy("year").pivot("course").agg(functions.sum("earnings")).show()
        +----+-----+------+
        |year| Java|dotNET|
        +----+-----+------+
        |2012|20000| 15000|
        |2013|30000| 48000|
        +----+-----+------+
        >>> df4.groupBy("year").pivot("course", ["dotNET", "PHP"]).agg(functions.sum("earnings")).show()
        +----+------+----+
        |year|dotNET| PHP|
        +----+------+----+
        |2012| 15000|null|
        |2013| 48000|null|
        +----+------+----+
        >>> df4.groupBy("year").pivot("course").agg(functions.sum("earnings"), functions.avg("earnings")).show()
        +----+------------------+------------------+--------------------+--------------------+
        |year|Java_sum(earnings)|Java_avg(earnings)|dotNET_sum(earnings)|dotNET_avg(earnings)|
        +----+------------------+------------------+--------------------+--------------------+
        |2012|             20000|           20000.0|               15000|              7500.0|
        |2013|             30000|           30000.0|               48000|             48000.0|
        +----+------------------+------------------+--------------------+--------------------+
        """
        jgd = self._jgd.pivot(functions.parse(pivot_col), values)
        return GroupedData(jgd, self._df)

    # pylint: disable=W0511
    # todo: implement apply()
    # def apply(self, udf):
    #     # Columns are special because hasattr always return True
    #     if isinstance(udf, Column) or not hasattr(udf, 'func') \
    #             or udf.evalType != PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF:
    #         raise ValueError("Invalid udf: the udf argument must be a pandas_udf of type "
    #                          "GROUPED_MAP.")
    #     df = self._df
    #     udf_column = udf(*[df[col] for col in df.columns])
    #     jdf = self._jgd.flatMapGroupsInPandas(udf_column._jc.expr())
    #     return DataFrame(jdf, self.sql_ctx)



from ..utils import IllegalArgumentException


def to_option_stored_value(value):
    if value is None:
        return None
    if isinstance(value, bool):
        return str(value).lower()
    return str(value)


class OptionUtils:
    def _set_opts(self, schema=None, **options):
        """
        Set named options (filter out those the value is None)
        """
        if schema is not None:
            self.schema(schema)
        for k, v in options.items():
            if v is not None:
                self.option(k, v)

    def option(self, key, value):
        raise NotImplementedError

    def schema(self, schema):
        # By default OptionUtils subclass do not support schema
        raise IllegalArgumentException(
            f"schema is not a valid argument for {self.__class__}"
        )



def resolve_column(col, row, schema, allow_generator=True):
    """
    Return the list of column names corresponding to a column value and a schema and:
    If allow generator is False, a list of values corresponding to a row
    If allow generator is True, a list of list of values, each list correspond to a row
    """
    output_cols = [field.name for field in col.output_fields(schema)]

    output_values = col.eval(row, schema)

    if not allow_generator and col.may_output_multiple_rows:
        raise Exception("Generators are not supported when it's nested in expressions,"
                        f" but got: {col}")

    if not col.may_output_multiple_rows:
        output_values = [output_values]
        if not col.may_output_multiple_cols:
            output_values = [output_values]

    return output_cols, output_values



from ._row import Row
from .column import Column
from .context import SQLContext  # , HiveContext, UDFRegistration
from .dataframe import DataFrame, DataFrameNaFunctions, DataFrameStatFunctions
from .group import GroupedData
from .readwriter import DataFrameReader, DataFrameWriter
from .session import SparkSession

# from pyspark.sql.catalog import Catalog
# from pyspark.sql.window import Window, WindowSpec
# from pyspark.sql.pandas.group_ops import PandasCogroupedOps

__all__ = [
    'SparkSession', 'SQLContext',  # 'HiveContext', 'UDFRegistration',
    'DataFrame', 'GroupedData', 'Column', 'Row',  # 'Catalog',
    'DataFrameNaFunctions', 'DataFrameStatFunctions',  # 'Window', 'WindowSpec',
    'DataFrameReader', 'DataFrameWriter',  # 'PandasCogroupedOps'
]
